/*
 *----------------------------------------------------------------------
 *
 * HostIF_ReleaseAnonPage --
 *
 *      The page with given mpn is unlocked and released. The page
 *      is put into a cache of released pages. Released pages can be 
 *      recycled later by calling HostIF_GetRecycledPage.
 *
 * Results:
 *      PAGE_UNLOCK_NO_ERROR
 *
 *----------------------------------------------------------------------
 */

int
HostIF_ReleaseAnonPage(VMDriver *vm, MPN mpn)
{
   MemTrackEntry *e;

   e = MemTrack_LookupMPN(vm->memtracker, mpn);
   if (e != NULL) {
      ASSERT(e->mpn == mpn);

      HOST_UNLOCK_PFN(vm, mpn);
      e->mpn = 0;

      MemTrack_RecyclePage(vm->memtracker, e);
   }
   return PAGE_UNLOCK_NO_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * HostIF_GetRecycledPage --
 *
 *      Returns the VPN of a previously released anonymous page, if
 *      one is available.
 *
 * Results:
 *      VPN of the page if one is available, INVALID_VPN otherwise
 *
 *----------------------------------------------------------------------
 */

VPN
HostIF_GetRecycledPage(VMDriver *vm)
{
   MemTrackEntry *e;

   e = MemTrack_GetRecycledPage(vm->memtracker);
   if (e != NULL) {
      return e->vpn;
   } else {
      return INVALID_VPN;
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * HostIF_UserToDriverPtr 
 *
 *    Convert a user virtual address into a kernel virtual address to access
 *    the same physical memory address --hpreg
 *
 * Results:
 *    The kernel virtual address on success
 *    NULL on failure
 *
 * Side effects:
 *    XXX Use 1 per-VM storage location. So for example, we leak kernel memory
 *        when we run a SMP VM on a host with more than 850 MB of RAM --hpreg
 *
 *-----------------------------------------------------------------------------
 */

void *
HostIF_UserToDriverPtr(VMDriver *vm, // IN
                       VA32 uvAddr) // IN
{
   struct page *page;

   if (HostIFGetUserPage((void *)uvAddr, &page)) {
      return NULL;
   }

   HostIF_GlobalLock(17);
   if (vm->vmhost->crosspagePagesCount >= MAX_INITBLOCK_CPUS) {
      HostIF_GlobalUnlock(17);
      put_page(page);
      return NULL;
   }
   vm->vmhost->crosspagePages[vm->vmhost->crosspagePagesCount++] = page;
   HostIF_GlobalUnlock(17);

   return (char *)kmap(page) + (uvAddr & (PAGE_SIZE - 1));
}


/*
 *-----------------------------------------------------------------------------
 *
 * __mod_user_asm --
 *
 *    Mark page dirty. Does not check for kernel address or for broken
 *    write protect on i386.
 *
 * Results:
 *    err unmodified on success
 *    err set to errret on failure.
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

#define __mod_user_asm(addr, err, errret)	\
   asm volatile (				\
      "1:    lock; orb $0,%1\n"			\
      "2:\n"					\
      ".section .fixup,\"ax\"\n"		\
      "3:    movl %2,%0\n"			\
      "      jmp 2b\n"				\
      ".previous\n"				\
      ".section __ex_table,\"a\"\n"		\
      "      .align 4\n"			\
      "      .long 1b,3b\n"			\
      ".previous"				\
      : "=r"(err)				\
      : "m"(__m(addr)), "i"(errret), "0"(err))


/*
 *-----------------------------------------------------------------------------
 *
 * mod_user_check --
 *
 *    Mark page dirty.
 *
 * Results:
 *    0 on success
 *    -EFAULT on failure - address not mapped, or read-only, or not user address
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

#define mod_user_check(addr)				\
({							\
   long __mu_err = -EFAULT;				\
   void *__mu_addr = (addr);				\
   if (access_ok(VERIFY_WRITE, __mu_addr, 1)) {		\
      __mu_err = 0;					\
      __mod_user_asm(__mu_addr, __mu_err, -EFAULT);	\
   }							\
   __mu_err;						\
})


/*
 *-----------------------------------------------------------------------------
 *
 * HostIF_MarkPageDirty --
 *
 *    Mark page dirty.
 *
 * Results:
 *    0 on success
 *    != 0 on failure.
 *
 * Side effects:
 *    Complain loudly if global VM lock is held.
 *
 *-----------------------------------------------------------------------------
 */

int
HostIF_MarkPageDirty(void *addr) // IN/OUT
{
#ifdef MODULE_X86_64
   /* XXX not implemented on x86-64. */
   return 0;
#elif defined(CONFIG_X86_4G) || defined(CONFIG_X86_UACCESS_INDIRECT)
   HostIFGetUserPage(addr, NULL);
   return 0;
#else
   return mod_user_check(addr);
#endif
}


/*
 *----------------------------------------------------------------------
 *
 * HostIF_IsMPNLocked
 *
 *      Return entryPtr if MPN is locked into memory
 *
 * Results:
 *      0 if sucessful and non-zero or error.
 *
 * Side effects:
 *      None.
 *----------------------------------------------------------------------
 */

static void *
FindFunc(void *arg, MemTrackEntry *entryPtr)
{
   MPN mpn = (MPN) arg;
   
   if (entryPtr->mpn == mpn) {
      return entryPtr;
   }
   return NULL;
}

int
HostIF_IsMPNLocked(VMDriver *vm, MPN mpn)
{
  MemTrackEntry *entry;

  entry = MemTrack_Scan(vm->memtracker,vm, FindFunc);
  return (entry != NULL);
}


static void *
CheckFunc(void *arg,               // IN
          MemTrackEntry *entryPtr) // IN
{
   VMDriver *vm = (VMDriver *)arg;
   MPN mpn;
   
   if (entryPtr->mpn == 0) {
      return NULL;
   }

   mpn = HostifVa2Mpn(VPN_2_VA(entryPtr->vpn));
   if (mpn == 0) { 
      /* Unmapped */
      return entryPtr;
   }

   if (entryPtr->mpn != mpn) {
      Warning("CheckFunc vpn 0x%x mpn 0x%x and not 0x%x \n", 
              entryPtr->vpn, mpn, entryPtr->mpn);
      vm->checkFuncFailed = TRUE;
   }
    
   return NULL;
}


/*
 *-----------------------------------------------------------------------------
 *
 * HostIF_CheckMemory
 *
 *      Make sure memory locking is OK.
 *
 *      This function runs as part of the application process, as such
 *      HostifVa2Mpn is fine
 *
 * Results:
 *      None
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VA
HostIF_CheckMemory(VMDriver *vm) // IN
{
   MemTrackEntry *entryPtr;
   
   vm->checkFuncFailed = FALSE;
   entryPtr = MemTrack_Scan(vm->memtracker, vm, CheckFunc);
   
   if (vm->checkFuncFailed) {
      return 1;
   }

   if (entryPtr) {
      return VPN_2_VA(entryPtr->vpn);
   }

   return 0;
}
